/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "NCanUtils.h"
#include "KCanServer.h"
#include "XmlParser.h"

using namespace NCanUtils;


void SetLogFilename(const char * Filename, int LogTime, char * LogFile)
{
  int Day, Month, Year;
  GetDateFromDays(LogTime, Day, Month, Year);
  
  sprintf(LogFile, "%s_%4.4d%2.2d%2.2d", Filename, Year, Month, Day);
  
  printf("log file: %s\n", LogFile);
}

#if defined(__MYSQL__)
  #include "KMySql.h"
#endif


int main(int argc, char* argv[])
{
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  const char * av1 = argc > 1 ? argv[1] : NULL;
  const char * av2 = argc > 2 ? argv[2] : NULL;
  const char * av3 = argc > 3 ? argv[3] : NULL;
  const char * av4 = argc > 4 ? argv[4] : NULL;
  const char * can_device = av1;
  
  char Line[256];
  char LogFile[1024];
  int  LogTime;
  bool csv   = av4 && !strcmp(av4, "csv"); // comma separated values
  bool use_mysql = false;
#if defined(__MYSQL__)
  use_mysql = av4 && !strcmp(av4, "mysql");
#endif
  bool DoLog = (av2 && !use_mysql);
  int  SampleTime = 10;
  
  printf("can-bus data logger\n"
         "copyright (c) 2014 Jürg Müller, CH-5524\n\n");
  
  KParser Parser;
  KXmlNode Root;
  
  bool use_xml = false;
  if (argc == 2)
  {
    use_xml = strlen(av1) > 4 &&
              !strcmp(av1, ".xml");
    if (use_xml)
    {
      if (!Parser.ParseFile(av1, &Root))
        return -1;
    } else
    if (strncmp(av1, "can", 3))
    {
      printf("first parameter must begin with \"can\" or end with \".xml\"\n\n");
      argc = 1;
    }
  }
  if (argc <= 1 && !use_xml)
  {
    printf("usage:\n"
#if defined(__MYSQL__)
           "can_logger <can dev> [<filename> | <mysql host>] <sample time in sec> [ csv | mysql ]\n  or\n"
           "can_logger <file.xml>\n\n"
           "example: can_logger can0 192.168.1.137 10 mysql\n"
           "or       can_logger can1 canlog 3600 csv\n\n");
#else
           "can_logger <can dev> [ <filename>  [ <sample time in sec> [ csv ] ] ]\n\n"
           "example: can_logger can1 canlog 3600 csv\n\n");
#endif
  }
  
#if defined(__MYSQL__)
  KMySql MySql;

  if (use_xml)
  {
    use_mysql = Root.Find("mysql") != NULL;
    if (use_mysql &&
        !MySql.GetXmlParams(&Root))
      return -1;
  } else
  if (use_mysql)
    strcpy(MySql.Host, av2);
#endif
  
  if (use_xml)
  {
    can_device = Root.FindNodeAttr("device", "Name");
    av3 = Root.FindNodeAttr("sleep", "Name");
    if (!use_mysql)
      av2 = Root.FindNodeAttr("log_file", "Name"); 
    SampleTime = 104;
  }
  KCanServer Can;
  KCanFrame Frame;
  
  int sec;
  if (av3 &&
      NCanUtils::GetInt(av3, sec) &&
      !*av3 &&
      5 <= sec && sec <= 24*3600)
    SampleTime = sec;
  
  if (DoLog)
  {
    ArtTime(Frame.TimeStampDay, Frame.TimeStampMs);
    LogTime = Frame.TimeStampDay;
    SetLogFilename(av2, LogTime, LogFile);
  }
  
  Can.Init(can_device);
  
  printf("can dev: %s\n", Can.GetDev());
  if (argc > 2)
  {
    if (!DoLog)
      printf("mysql host: %s\n", MySql.Host);
  }
  printf("sample time: %d sec\n", SampleTime);
  if (csv)
    printf("use comma separated values\n");
            
  printf("\n");
  
  Can.Resume();
  
#if defined(__MYSQL__)
  if (use_mysql)
  {
    strcpy(MySql.Host, av2);
    
    // test connection
    if (!MySql.Open())
    {
      Can.Halt();
      
      return -1;
    }
    MySql.Close();
  }
#endif
  
  while (!Can.Terminated())
  {
    for (int i = 0; i < 10*SampleTime && !Can.RingBufferFull(); i++)
      SleepMs(100);
    
    FILE * log = NULL;
    bool read = false;
    while (Can.GetFrame(Frame))
    {
#if defined(__MYSQL__)
      if (use_mysql)
      {
        if (!MySql.Opened())
          MySql.Open();
        if (!MySql.Opened())
        {
          Can.Terminate();
          break;
        }
        MySql.LogFrame(Frame);
      }
#endif
      read = true;
      FormatFrame(Frame, csv, Line);
      if (DoLog)
      {
        // new day -> use new file
        if (Frame.TimeStampDay != LogTime)
        {
          LogTime = Frame.TimeStampDay;
          SetLogFilename(av2, LogTime, LogFile);
          
          if (log)
            fclose(log);
          log = NULL;
        }
        if (!log)
        {
          log = fopen(LogFile, "ab+");
          if (!log)
          {
            printf("log file \"%s\" not opened\n", LogFile);
            Can.Terminate();
            break;
          }
        }
        if (log)
          fwrite(Line, 1, strlen(Line), log);
      }
      // in DoLog or mysql mode: show only last frame
      if (Frame.Id & 0x60000000 || // RTR/ERR flags
          !(DoLog || use_mysql) || trace)
      {
        printf("%s", Line);
        read = false;
      }
    }
    if (log)
      fclose(log);
    if (read)
      printf("%s", Line);
#if defined(__MYSQL__)
    if (use_mysql)
    {
      if (MySql.Opened())
        MySql.Close();
    }
#endif
  }
  char stream[64];
  
  NCanUtils::SleepMs(1000);
  
  ArtTime(Frame.TimeStampDay, Frame.TimeStampMs);
  PrintTime(Frame, stream, false);
  printf("%s terminated\n", stream);
  
  return 0;
}



