/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__LINUX__) 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(__MYSQL__)
  #include "KMySql.h"
#endif

#include "KCanElster.h"

#include "ElsterTable.inc"

int GetElsterTableIndex(unsigned short Index)
{
  return ElsterTabIndex[Index];
}

const char * GetElsterTableName(unsigned short Index)
{
  int ind = GetElsterTableIndex(Index);
  if (ind >= 0)
    return ElsterTable[ind].Name;
  else
    return "";
}

ElsterType GetElsterType(const char * str)
{
  ElsterType type = et_default;
  if (!str)
    return type;
  
  if (!strcmp(str, "et_dec_val"))
    type = et_dec_val;
  else
  if (!strcmp(str, "et_byte"))
    type = et_byte;
  else
  if (!strcmp(str, "et_little_endian"))
    type = et_little_endian;
  else
  if (!strcmp(str, "et_zeit"))
    type = et_zeit;
  else
  if (!strcmp(str, "et_datum"))
    type = et_dec_val;

  return type;
}

////////////////////////////////////////////////////////////////////////////////

void KSniffFrame::SetValue(unsigned Value)
{
  if (3 <= Len && Len <= 5)
  {
    Data[Len] = (unsigned char) (Value >> 8);
    Data[Len+1] = (unsigned char) (Value & 0xff);  
  }
}

unsigned KSniffFrame::GetValue() const
{
  if (3 <= Len && Len <= 5)
  {
    return 256*Data[Len] + Data[Len + 1];
  }
  return 0;
}


////////////////////////////////////////////////////////////////////////////////

KCanElster::KCanElster()
{
  Trace = false;
  SendCanId = 0x700;
  
  for (int i = 0; i <= (int) High(Sniffed); i++)
    Sniffed[i] = NULL;
}

KCanElster::~KCanElster()
{
  for (int i = 0; i <= (int) High(Sniffed); i++)
  {
    while (Sniffed[i])
    {
      KSniffFrame * SF = Sniffed[i]->Next;
      
      delete Sniffed[i];
      Sniffed[i] = SF;
    }
  }
}

bool KCanElster::Init(const char * CanDev, unsigned short DefaultId)
{
  Trace = false;
  SendCanId = DefaultId;
  if (!KCanServer::Init(CanDev))
    return false;
  
  SendDriver.Init(CanDev);
  if (!SendDriver.Connect())
    return false;

  return true;
}

bool KCanElster::AddToSniffed(const KCanFrame & Frame)
{
  unsigned char Index = Frame.Data[0];
  
  if ((Frame.Len != 3 && Frame.Len != 5) ||
      (Index & 0xf) == 1)
    return false;
  
  if (SearchSniffFrame(Frame))
    return false;
  
  KSniffFrame * SF = new KSniffFrame;
  
  SF->Len = Frame.Len;
  for (int i = 0; i < Frame.Len; i++)
    SF->Data[i] = Frame.Data[i];
  
  SF->SetValue(0x8000); // Wert ist ungültig
  
  SF->Next = Sniffed[Frame.Data[0]];
  Sniffed[Frame.Data[0]] = SF;
  
  return true;  
}

KSniffFrame * KCanElster::SearchSniffFrame(const KCanFrame & Frame)
{
  KSniffFrame * SF = Sniffed[Frame.Data[0]];
  
  while (SF)
  {
    if (Frame.Len >= SF->Len)
    {
      for (int i = SF->Len-1; i >= 0; i--)
      {
        if (SF->Data[i] != Frame.Data[i])
          break;
        
        if (i <= 0)
          return SF;
      }
    }
    SF = SF->Next;
  }
  return NULL;
}

void KCanElster::PrintFrame(const KCanFrame & Frame)
{
  char Buffer[256];
  
  NCanUtils::FormatFrame(Frame, false, Buffer);
  printf("%s", Buffer);
}

bool KCanElster::Send_Frame()
{
  bool Ok = SendDriver.SendData(SendFrame);
  
  if (Trace)
    PrintFrame(SendFrame);
  
  return Ok;
}

bool KCanElster::SniffFrame()
{
  bool Ok = GetFrame(RecvFrame);
  
  if (Ok)
  {
    KSniffFrame * SF = SearchSniffFrame(RecvFrame);
    if (SF &&
        RecvFrame.Len >= SF->Len + 2)
    {
      int len = SF->Len;
      SF->Data[len] = RecvFrame.Data[len];
      SF->Data[len+1] = RecvFrame.Data[len+1];
      SF->TimeStampDay = RecvFrame.TimeStampDay;
      SF->TimeStampMs = RecvFrame.TimeStampMs;
    }
  }
  return Ok;
}

void KCanElster::EmptyServer()
{
  bool Ok;
  do
  {
    Ok = SniffFrame();
    if (Ok && Trace &&
        SearchSniffFrame(RecvFrame))
      PrintFrame(RecvFrame);
  } while (Ok);
}

bool KCanElster::Get_Frame()
{
  bool Ok = SniffFrame();
  
  if (Ok && Trace)
    PrintFrame(RecvFrame);
  
  return Ok;
}

bool KCanElster::Send(unsigned count, bool WaitForAnswer)
{
  for (int i = 0; (unsigned) i < count; i++)  // 3 Versuche
  {
    EmptyServer();
    
    if (Send_Frame())
    {
      NCanUtils::SleepMs(10);
      while (Get_Frame())
      {
        if (RecvFrame.EqualData(SendFrame))
        {
          if (!WaitForAnswer)
            return true;
          
          for (int k = 0; k < 100; k++)
          {
            NCanUtils::SleepMs(1);
            if (Get_Frame() &&
                RecvFrame.IsAnswer(SendFrame))
            {
              return true;
            }
          }
        }
      }
    }
    if ((unsigned)(i+1) < count)
      NCanUtils::SleepMs(1000);
  }
  
  return false;
}

void KCanElster::InitSendFrame(unsigned short receiver_id, unsigned short TelegrammId)
{
  unsigned char address = (receiver_id & 0x780) / 8;
  SendFrame.Init();
  SendFrame.Id = SendCanId;
  SendFrame.Len = 5;
  SendFrame.Data[0] = (unsigned char)((address & 0xf0) + 1);
  SendFrame.Data[1] = (receiver_id & 7);
  SendFrame.Data[2] = 0xfa;
  SendFrame.Data[3] = (unsigned char)(TelegrammId >> 8);
  SendFrame.Data[4] = (unsigned char)(TelegrammId & 0xff);
}

bool KCanElster::GetValue(unsigned short receiver_id, unsigned short TelegrammId, unsigned short & Value)
{
  InitSendFrame(receiver_id, TelegrammId);
  
  if (Send() && RecvFrame.Len == 7)
  {
    Value = RecvFrame.Value();
    return true;
  }
  
  return false;
}

bool KCanElster::SetValue(unsigned short receiver_id, unsigned short TelegrammId, unsigned short Value)
{
  InitSendFrame(receiver_id, TelegrammId);
  
  SendFrame.Len = 7;
  SendFrame.Data[0]++;
  SendFrame.Data[5] = (unsigned char)(Value >> 8);
  SendFrame.Data[6] = (unsigned char)(Value & 0xff);
  
  if (!Send(1, false))
    return false;
  
  NCanUtils::SleepMs(50);
  SendFrame.Data[0]--;
  if (!Send())
    return false;

  bool Ok = RecvFrame.Changed(SendFrame);
  
  SendFrame.Data[0]++;
  
  return Ok;
}

void KCanElster::Scanner(unsigned short receiver_id)
{
  unsigned short Value;
  char Buffer[256];
  
  for (unsigned Pattern = 0; Pattern < (receiver_id & 7) ? 0x07ff : 0x1fff; Pattern++)
  {
    if (GetValue(receiver_id, Pattern, Value) &&
        Value != 0x8000 &&
        (Value || !(receiver_id & 7)))
    {
      int ind = GetElsterTableIndex(Pattern);
      if (ind >= 0 ||
          Value != 0x8000)
      {
        sprintf(Buffer, "%4.4x:  %4.4x", Pattern, Value);
        if (ind >= 0)
        {
          if (Value != 0x8000)
          {
            char Val[32];
          
            SetValueType(Val, ElsterTable[ind].Type, Value);
            sprintf(Buffer + strlen(Buffer), "  (%s  %s)", ElsterTable[ind].Name, Val);
          } else
            sprintf(Buffer + strlen(Buffer), "  (%s)", ElsterTable[ind].Name);
        }
        printf("%s\n", Buffer);
      }
    }
  }  
}

#endif

#if defined(__SCAN__)

void PrintComfortSoftFormat(const KCanFrame & Frame)
{
  char str[128];
  KCanFrame Frame1;
  KComfortFrame cf;
  KCan232Frame can232;
  
  cf.GetCanFrame(Frame);
  cf.SetToString(str);
  can232.GetCanFrame(Frame);
  printf("ComfortSoft: %s   can232: %s\n", str, can232.msg);
  cf.SetCanFrame(Frame1);
  
  bool Equal = Frame.Id == Frame1.Id;
  for (int i = 0; i < Frame.Len; i++)
    if (Frame.Data[i] != Frame1.Data[i])
      Equal = false;
  if (!Equal)
  {
    NCanUtils::FormatFrame(Frame1, false, str);
    printf("invalid back translated telegram: %s\n", str);
  }
}

int main(int argc, char* argv[])
{
  KCanElster Can;
  
  printf("elster-kromschröder can-bus address scanner and test utility\n"
         "copyright (c) 2014 Jürg Müller, CH-5524\n\n");
  
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  bool params_ok = argc == 4;
  bool scan_only = false;
  bool write = false;
  unsigned short Index = 0;
  unsigned short Value = 0;
  const char * Line;
  const char * last_pos;  
  unsigned sender_can_id = 0;
  unsigned scan_can_id = 0;
  
  if (params_ok)
  {
    Line = argv[2];
    sender_can_id = NCanUtils::GetHex(Line);
    switch (sender_can_id & ~3)
    {
      case 0x680:
      case 0x700:
      case 0x780:
        break;
      default:
        params_ok = false;
        break;
    }
  }
  
  if (params_ok)
  {
    Line = argv[3];
    last_pos = Line;
    // 000.0000 oder 000.0000.0000
    scan_can_id = NCanUtils::GetHex(Line);
    if (Line - last_pos != 3 ||
        (scan_can_id < 0x100 && scan_can_id > 7) ||
        scan_can_id > 0x687 ||
        (scan_can_id & 0x078))
    {
      printf("\n%x is not a legal can id!\n\n", scan_can_id);
      params_ok = false;
    }
  
    last_pos = Line;
    scan_only = *Line != '.';
    if (!scan_only)
    {
      Index = NCanUtils::GetHex(++Line);
      if (Line - last_pos != 5)
        params_ok = false;
    }
    last_pos = Line;
    write = *Line == '.';
    if (write)
    {
      Value = NCanUtils::GetHex(++Line);
      if (Line - last_pos != 5 ||
          sender_can_id == 0)
        params_ok = false;
    }
  }
  if (!params_ok)
  {
    printf("use scanner\n"
           "can_scan <can dev> <sender can id> <scan can id> [ \"trace\" ]\n"
           "  or read/change value\n"
           "can_scan <can dev> <sender can id> <receiver can id> \".\" <elster index> [ \".\" <new value> ] [ \"trace\" ]\n\n"
           "example: can_scan can0 780 180           (scan all elster indices from 0000 to 1fff)\n"
           "or       can_scan can0 780 180.0126      (read minutes at elster index 0126)\n"
           "or       can_scan can0 780 180.0126.0f00 (set minutes to 15)\n\n"
           "allowed sender can id: 680 or 700 or 780\n\n"
           "BE CAREFUL WITH THIS TOOL !!!\n\n");
    return 0;
  }

  if (!Can.Init(argv[1], sender_can_id))
  {
    printf("%s not connected\n", argv[1]);
    return -1;
  }
  
  Can.Trace = trace;
  Can.Resume();
   
  if (scan_only)
  {
    Can.Scanner(scan_can_id);
  } else {
    bool Ok;
  
    if (write)
      Ok = Can.SetValue(scan_can_id, Index, Value);
    else 
      Ok = Can.GetValue(scan_can_id, Index, Value);
    
    if (Ok && trace)
    {
      printf("\nsend ");
      PrintComfortSoftFormat(Can.SendFrame);
      printf("recv ");
      PrintComfortSoftFormat(Can.RecvFrame);
      printf("\n");
      
    }
    
    if (!Ok)
    {
      if (write)
        printf("value not set\n");
      else
        printf("no value read\n");
    } else {
      if (write)
        printf("value set\n");
      
      printf("value: %4.4x", Value);

      int ind = GetElsterTableIndex(Index);
      if (ind >= 0)
      {
        char Val[32];
        
        SetValueType(Val, ElsterTable[ind].Type, Value);
        printf("  (%s  %s)", ElsterTable[ind].Name, Val);
      }
      printf("\n");    
    }
  }
  Can.Halt();
  
  return 0;  
}

#endif

#if defined(__VZ__)

#include "XmlParser.h"
#include "KTcpCanClient.h"

static char HttpBuffer[12400];

int main(int argc, char* argv[])
{ 
  printf("Volkszaehler can-bus logger\n"
         "copyright (c) 2014 Jürg Müller, CH-5524\n\n");

  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  if (argc != 2)
  {
    printf("usage:\n"
           "can_vz <xml configuration> [ \"trace\" ]\n"
           "example: can_vz VZ_config.xml\n\n");
    return 0;
  }
  
  KParser Parser;
  KXmlNode Root;
  const KXmlNode * Node;
  char buffer[1024];
  
  if (!Parser.ParseFile(argv[1], &Root))
    return -1;
  
  const char * device = Root.FindNodeAttr("device", "Name");
  const char * template_ = Root.FindNodeAttr("template", "Name");
  const char * ip_address = Root.FindNodeAttr("ip_address", "Name");
  const char * secs = Root.FindNodeAttr("sleep", "Name");
#if defined(__MYSQL__)
  bool use_mysql = Root.Find("mysql") != NULL;
  KMySql MySql;
#else
  bool use_mysql = false;
#endif
  int seconds = 104;
  
  if (!use_mysql &&
      (!template_ || strlen(template_) == 0))
  {
    printf("template is missing\n");
    return -1;
  }
  if (!use_mysql && (!ip_address || strlen(ip_address) == 0))
  {
    printf("IP address is missing\n");
    return -1;
  }
  
#if defined(__MYSQL__)
  if (use_mysql &&
      !MySql.GetXmlParams(&Root))
    return -1;
#endif
  
  if (secs)
  {
    int sec;
    if (NCanUtils::GetInt(secs, sec) &&
        !*secs &&
        60 <= sec && sec <= 3600)
      seconds = sec;
  }
  if (trace)
  {
    printf("device: %s\n", device);
    printf("sleep: %d secs\n", seconds);
    if (!use_mysql)
      printf("IP address: %s\n", ip_address);
    printf("\n");
  }
  
  // cannel check
  int cnt = 1;
  Node = NULL;
  while (Root.GetNextNode(Node))
  {
    if (strlen(Node->mName) == 0)
      return -1;
    
    if (!strcmp(Node->mName, "channel"))
    {
      const char * can_msg = Node->FindAttr("can_msg");
      const char * uuid = Node->FindAttr("uuid");
      const char * query = Node->FindAttr("query");
      const char * name = Node->FindAttr("Name");
      
      if (!name || strlen(name) == 0)
      {
        printf("\"Name\" is missing in channel %d\n", cnt);
        return -1;
      }
      if (use_mysql)
      {
        if (!query || strlen(query) == 0)
        {
          printf("query is missing in channel %d (%s)\n", cnt, name);
          return -1;
        }
        if (!strstr(query, "${value}"))
        {
          printf("${value} in query is missing %s (%s)\n", query, name);
          return -1;
        }
      } else {
        if (!uuid || strlen(uuid) == 0)
        {
          printf("uuid is missing in channel %d (%s)\n", cnt, name);
          return -1;
        }
      }
      if (!can_msg || strlen(can_msg) == 0)
      {
        printf("can_msg is missing in channel %d (%s)\n", cnt, name);
        return -1;
      } else {
        KCanFrame Frame;
        
        if (!Frame.GetDataFromStream(can_msg))
        {
          printf("invalid can_msg(%s): \"%s\"\n", name, can_msg);
          return -1;
        }
      }
      cnt++;
    }
  }  
  
  KCanElster Can;
  bool added = false;
  
  Node = NULL;
  while (Root.GetNextNode(Node))
  {
    if (!strcmp(Node->mName, "channel"))
    {
      KCanFrame Frame;
      const char * can_msg = Node->FindAttr("can_msg");
      
      Frame.GetDataFromStream(can_msg);
      if (Can.SearchSniffFrame(Frame))
      {
        printf("can_msg \"%s\" already exists\n", can_msg);
      }
      if (Can.AddToSniffed(Frame) && trace)
      {
        printf("frame can_msg added: %s\n", can_msg);
        added = true;
      }
    }
  }
  if (added)
    printf("\n");
  
  KTcpCanClient TcpClient;
  time_t timer;
  time_t new_timer;

  TcpClient.Trace = trace;
  TcpClient.SetBlockMode(bmBlocking);
  TcpClient.SetRemoteHost(ip_address);
  TcpClient.SetRemotePort(80);
  
  if (!Can.Init(device, 0x700))
  {
    printf("%s not connected\n", Can.GetDev());
   
    Can.Halt();
    
    return -1;
  }

  Can.Trace = trace;
  Can.Resume();
  
  NCanUtils::SleepMs(10000);
  while (!Can.Terminated())
  {
    Node = NULL;
    time(&timer);
    while (Root.GetNextNode(Node))
    {
      if (!strcmp(Node->mName, "channel"))
      {
        const char * can_msg = Node->FindAttr("can_msg");
        
        const_cast<KXmlNode *>(Node)->can_value = 0x8000;

        Can.SendFrame.Init();
        Can.SendFrame.Id = Can.SendCanId;
        Can.SendFrame.GetDataFromStream(can_msg);
        if ((Can.SendFrame.Data[0] & 0xf) != 1)
        {
          Can.EmptyServer();
          
          KSniffFrame * SF = Can.SearchSniffFrame(Can.SendFrame);
          if (SF)
          {
            const_cast<KXmlNode *>(Node)->can_value = SF->GetValue();
            const_cast<KXmlNode *>(Node)->time_days = SF->TimeStampDay;
            const_cast<KXmlNode *>(Node)->time_ms = SF->TimeStampMs;
  
            SF->SetValue(0x8000);
          }
        } else
        if (Can.Send())
        {
          const_cast<KXmlNode *>(Node)->can_value = Can.RecvFrame.Value();
          const_cast<KXmlNode *>(Node)->time_days = Can.RecvFrame.TimeStampDay;
          const_cast<KXmlNode *>(Node)->time_ms = Can.RecvFrame.TimeStampMs;
        }
      }
    }
    
  #if defined(__MYSQL__)
    if (use_mysql)
    {
      MySql.Open();
    }
  #endif
    Node = NULL;
    while (Root.GetNextNode(Node))
    {
      if (!strcmp(Node->mName, "channel"))
      {
        if (Node->can_value != 0x8000)
        {
          const char * format = Node->FindAttr("format");
          ElsterType type = GetElsterType(format);
          char Value[64];
          SetValueType(Value, type, Node->can_value);
          
        #if defined(__MYSQL__)
          if (use_mysql)
          {
            char res[1024];
            const char * query = Node->FindAttr("query");
            
            res[0] = 0;
            while (query && *query)
            {
              const char * pos_val = strstr(query, "${value}");
              const char * pos_ts = strstr(query, "${ts}");
              const char * pos = !pos_ts ? pos_val : pos_ts;
     
              if (pos_ts && pos_val && pos_val < pos_ts)
                pos = pos_val;
              if (pos)
              {
                int n = (int)(pos - query);
                int len = (int) strlen(res);
                strncpy(res + len, query, n);
                res[len + n] = 0;
                query += n;
                if (pos == pos_val)
                {
                  strcat(res, Value);
                  query += 8;
                } else {
                  char time_buf[64];
            
                  NCanUtils::PrintDateTime(Node->time_days, Node->time_ms, time_buf);
                  //NCanUtils::PrintMs(Node->time_days, Node->time_ms, time_buf);
                  strcat(res, time_buf);
                  query += 5;
                }
              } else {
                strcat(res, query);
                query = NULL;
              }
            }
        
            if (!MySql.Query(res))
              printf("error in query %s\n", res);
            else
            if (trace)
              printf("query: %s\n", res);            
          } else
        #endif
          {
            const char * uuid = Node->FindAttr("uuid");
            
            sprintf(buffer, template_, uuid, Value);
            if (trace)
              printf("==> %s\n", buffer);
            if (!TcpClient.ReadHttpStream(HttpBuffer, High(HttpBuffer), buffer))
            {
              printf("TCP/IP server not connected\n");
              if (!trace)
                printf("==> %s\n", buffer);
            }
          }
          const_cast<KXmlNode *>(Node)->can_value = 0x8000;            
        } else
        if (trace)
        {
          const char * name = Node->FindAttr("Name");
          printf("\"%s\" has non or invalid value\n", name);
        }
      }
    }
    
  #if defined(__MYSQL__)
    if (use_mysql)
    {
      MySql.Close();
    }
  #endif
    
    time(&new_timer);
    int del = seconds - (int)(new_timer - timer);
    if (del < 45)
      del = 45;
    while (del > 0)
    {
      int d = del < 10 ? del : 10;
      
      NCanUtils::SleepMs(d*1000);
      del = del < 10 ? 0 : del - 10;
      
      Can.EmptyServer();
    }
  }
  
  return 0;
}

#endif

