/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */


#if defined(__LINUX__)
  #include <sys/types.h>
  #include <sys/time.h>
  #include <unistd.h>
#endif
#if defined(__WINDOWS__)
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "NCanUtils.h"

#if defined (__CAN__)
  unsigned gMainThreadId = 0;
  unsigned gProcessId = 0;
#endif

namespace NCanUtils
{
  
void SleepMs(int aMs)
{
#if defined(__WINDOWS__)
  Sleep(aMs);
#endif
#if defined(__LINUX__)
  usleep(1000*aMs);
#endif
}
  
void ArtTime(const struct timeval & Time, int & aDays, int & aMs)
{
  aMs = 1000*(Time.tv_sec % 86400) + Time.tv_usec/1000;
  aDays = (int) (Time.tv_sec / 86400);
  // 72 76 80 84 88 92 96 : 7 Schaltjahre
  aDays -= (2000-1970)*365 + (2000-1970)/4;
  if (aMs < 0)
  {
    aMs += 1000*86400;
    aDays--;
  }
}

#if defined(__LINUX__)

void ArtTime(int & aDays, int & aMs)
{
  struct timeval Time;

  if (!gettimeofday(&Time, NULL))  // seit 1. Jan. 1970
  {
    ArtTime(Time, aDays, aMs);
  } else {
    aMs = 0;
    aDays = 0;
  }
}

#endif

static const unsigned cMonthDays[13] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};
// jan feb mar  apr  mai  jun  jul  aug  sep  okt  nov  dez
//  31  28  31   30   31   30   31   31   30   31   30   31
void GetDateFromDays(int aDays, int & aDay, int & aMonth, int & aYear)
// aDays == 0 entspricht dem 1.1.2000 (aDay = 1, aMonth = 1, aYear = 2000)
{
  aMonth = 1;
  aDay = aDays;
  
  // immer 4 Jahre abzählen: 4*365 Tage + 1 Tag vom Schaltjahr
  aYear = aDay / (4*365 + 1);
  aDay -= aYear*(4*365 + 1);
  aYear = 4*aYear + 2000;
  if (aDay <= 0)
  {
    aDay += 4*365 + 1;
    aYear -= 4;
  }
  // das 1. Jahr (z.B. 2012) ist ein Schaltjahr, danach bis zu 3 normale Jahre
  if (aDay >= 366)
  {
    aYear++;
    aDay -= 366;
    
    while (aDay >= 365)
    {
      aYear++;
      aDay -= 365;
    }
  }
  aDay++; // der erste Tag im Monat ist der 1.
  while ((int) cMonthDays[aMonth] < aDay) 
  {
    if (aMonth == 2 && !(aYear & 3))
    {
      if ((int) cMonthDays[2]+1 >= aDay)  // 28. Feb. in einem Schaltjahr
        break;
      aDay--;    // für den 29. Feb. (ist nicht in cMonthDays enthalten)
    }
    aMonth++;
  }
  aDay -= cMonthDays[aMonth-1];
}

void PrintDayTime(int time_ms, char * OutStr)
{
  int Sec = time_ms / 1000;
  int Min = Sec / 60;
  int Hour = Min / 60;
  
  sprintf(OutStr, "%2.2d:%2.2d:%2.2d.%3.3d",
          Hour, Min % 60, Sec % 60, time_ms % 1000);
}
  
void PrintTime(int time_days, int time_ms, char * OutStr, bool csv)
{
  int Day, Month, Year;
  
  GetDateFromDays(time_days, Day, Month, Year);
  if (csv)
  {
    sprintf(OutStr, "'%4.4d-%2.2d-%2.2d',%d", Year, Month, Day, time_ms);
  } else {
    char day_str[32];
    
    PrintDayTime(time_ms, day_str);
    sprintf(OutStr, "%d.%d.%d %s",
            Day, Month, Year, day_str);
  }
}
  
void PrintTime(const KCanFrame & Frame, char * OutStr, bool csv)
{
  PrintTime(Frame.TimeStampDay, Frame.TimeStampMs, OutStr, csv);
}  

void PrintMs(int days, int ms, char * stream)
{
  long long time_stamp = days;
  
  time_stamp += 30*365 + 30 / 4; // Tage zwischen 1970 und 2000
  time_stamp *= 24*3600*1000;
  time_stamp += ms;
  
  // 120 + 6 + 365*14 + 4 = 5240
  // 30*365 + 7 = 10957
  // 16197
  // $ date -d "2014-05-06 05:17:25" +%s
  // 1399346245 / (24*3600) = 16196.13709
  // 0.13709 * 24 * 3600 = 11844.576
  // 11845 / 3600 = 3.2907 ==> verwendet GMT  (+ 1h + 1h Sommerz. ist effektive Zeit)
  // 1045 / 60 = 17.41..
  // 1045 - 60*17 = 25
  sprintf(stream, "%lld", time_stamp); // resulat stimmt nicht!
}
  
void PrintDateTime(int days, int ms, char * stream)
{
  // format: YYYY-MM-DD HH:MM:SS.mmm
  int Day, Month, Year;
  char day_str[32];
  
  GetDateFromDays(days, Day, Month, Year);
  PrintDayTime(ms, day_str);
  
  sprintf(stream, "%4.4d-%2.2d-%2.2d %s", Year, Month, Day, day_str);
}
  
void FormatFrame(const KCanFrame & Frame, bool csv, char * OutStr)
{
  char TimeStr[64];
  
  PrintTime(Frame, TimeStr, csv);
  if (csv)
  {
    sprintf(OutStr, "%u,%s,%u,%d", Frame.Counter, TimeStr,
            Frame.Id & 0xffffffff, Frame.Flags & 0xff);
    for (int i = 0; i < 8; i++)
    {
      strcat(OutStr, ",");
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%d", Frame.Data[i]);
      else
        strcat(OutStr, "NULL");
    }
  } else {
    if (!Frame.Counter && !Frame.TimeStampDay)
      sprintf(OutStr, "%8x %2.2x [%d] ", 
              Frame.Id & 0x1fffffff, Frame.Flags & 0xff, Frame.Len);
    else
      sprintf(OutStr, "%3u  %s  %8x %2.2x [%d] ", Frame.Counter, TimeStr,
              Frame.Id & 0x1fffffff, Frame.Flags & 0xff, Frame.Len);
    for (int i = 0; i < 8; i++)
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%2.2x ", Frame.Data[i]);
      else
        sprintf(OutStr + strlen(OutStr), "   ");
    strcat(OutStr, " ");
    for (int i = 0; i < 8; i++)
      if (i < Frame.Len)
        sprintf(OutStr + strlen(OutStr), "%c",
                ' ' <= Frame.Data[i] && Frame.Data[i] < 127 ? Frame.Data[i] : '.');
  }
  strcat(OutStr, "\n");
}
  
bool IsHexDigit(char aDigit)
{
  if (('0' <= aDigit && aDigit <= '9') ||
      ('a' <= aDigit && aDigit <= 'f') ||
			('A' <= aDigit && aDigit <= 'F'))
    return true;
  
  return false;
}

int GetHexDigit(char aDigit)
{
  if ('0' <= aDigit && aDigit <= '9')
    return (int) aDigit - '0';
  
  if ('A' <= aDigit && aDigit <= 'F')
    return (int) aDigit - ('A' - 10);
  
  if ('a' <= aDigit && aDigit <= 'f')
    return (int) aDigit - ('a' - 10);
  
  return -1;
}

unsigned GetHex(const char * & Line)
{
  if (!Line)
    return 0;
  
  unsigned Hex = 0;
  
  while (IsHexDigit(*Line))
  {
    Hex = 16*Hex + GetHexDigit(*Line++);
  }
  return Hex;
}
  
bool GetHexByte(const char * & Line, unsigned char & byte)
{
  if (!IsHexDigit(Line[0]) ||
      !IsHexDigit(Line[1]))
    return false;
  
  byte = 16*GetHexDigit(Line[0]) +
            GetHexDigit(Line[1]);
  Line += 2;
  
  return true;  
}
  
bool GetInt(const char * & Line, int & res)
{
  bool neg = *Line == '-';
  if (neg)
    Line++;
  bool Ok = ('0' <= *Line && *Line <= '9');
  
  res = 0;
  while ('0' <= *Line && *Line <= '9')
  {
    res = 10*res + (unsigned char)(*Line) - '0';
    Line++;
  }
  
  return Ok;
}
  
}

////////////////////////////////////////////////////////////////////////////////

KCanFrame::KCanFrame()
{
  Init();
}

void KCanFrame::Init()
{
  memset(this, 0, sizeof(*this));
}

bool KCanFrame::EqualData(const KCanFrame & Frame) const
{
  if (Len != Frame.Len)
    return false;
  
  for (int i = 0; i < Len; i++)
    if (Data[i] != Frame.Data[i])
      return false;
  
  return true;
}

bool KCanFrame::Changed(const KCanFrame & SendFrame) const
{
  if (Len < 5 || !IsAnswer(SendFrame))
    return false;
  
  return Value() == SendFrame.Value();
}

bool KCanFrame::IsAnswer(const KCanFrame & Frame) const
{
  // CAN-Ids vergleichen
  return (((Frame.Id / 8) & 0xf0) == (Data[0] & 0xf0)) &&
         ((Id / 8) & 0xf0) == (Frame.Data[0] & 0xf0);
}

bool KCanFrame::Supported() const
{
  if (Len < 5)
    return true;
  
  return Value() != 0x8000;
}

unsigned KCanFrame::Value() const
{
  if (Len < 2)
    return 0;
  
  return 256*Data[Len-2] + Data[Len-1];
}

bool KCanFrame::GetDataFromStream(const char * str)
{
  Len = 0;
  for ( ; Len < 8; Len++)
  {
    while (*str == ' ' || *str == '.')
      str++;
    
    if (!NCanUtils::GetHexByte(str, Data[Len]))
      break;
  }
  while (*str == ' ' || *str == '.')
    str++;
  
  return !*str || *str == '\n' || *str == '\r';
}


////////////////////////////////////////////////////////////////////////////////

// serielle Schnittstelle ComfortSoft
//
// send: 0d 00 0a 01 00 fa 01 99 00 00 01 ac    0d 00 09 01 00 fa 01 99 00 00 01 ac
// recv: 0a 00 0d 02 00 fa 01 99 f0 01 02 9e

// 0d.00.03.01.00.fa.01.99.00.00.01.a5
// coco: 0d 00 00 21 00 2f 00

static KComfortFrame InitFrame1 = {
  {0x0d, 0x00, 0x0d, 0x01, 0x00, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x26}
};

static KComfortFrame InitFrame2 = {
  {0x0d, 0x00, 0x00, 0x00, 0x00, 0xfa, 0x01, 0x5b, 0x00, 0x00, 0x01, 0x63}
};

unsigned KComfortFrame::GetCheckSum() const
{
  return 256*Data[10] + Data[11];
}

unsigned KComfortFrame::GetSum() const
{
  unsigned Sum = 0;
  for (int i = 0; i < 10; i++)
    Sum += Data[i];
  
  return Sum;
}

bool KComfortFrame::CheckSum() const
{
  return GetSum() == GetCheckSum();
}

void KComfortFrame::SetCheckSum()
{
  unsigned Sum = GetSum();
  
  Data[10] = (unsigned char)(Sum >> 8);
  Data[11] = (unsigned char)(Sum & 0xff);
}

bool KComfortFrame::SetCanFrame(KCanFrame & Frame) const
{
  Frame.Init();
  Frame.Len = 7;
  
  Frame.Id = (Data[0] << 7) + (Data[1] & 0x7);
  
  Frame.Data[0] = (Data[2] << 4) + (Data[3] & 0x0f);
  for (int i = 0; i < 6; i++)
    Frame.Data[i + 1] = Data[i + 4];
  
  return CheckSum();
}

bool KComfortFrame::GetCanFrame(const KCanFrame & Frame)
{
  Data[0] = (Frame.Id >> 7) & 0x0f;
  Data[1] = Frame.Id & 0x07;
  Data[2] = Frame.Data[0] >> 4;  // can-id 680
  Data[3] = Frame.Data[0] & 0x0f;
  Data[4] = Frame.Data[1];
  for (int i = 0; i < 5; i++)
    Data[i + 5] = Frame.Data[i + 2];
  
  SetCheckSum();
  
  return Frame.Len == 7;
}

void KComfortFrame::SetOk(bool Ok)
{
  for (int i = 0; i < 10; i++)
    Data[i] = Ok ? 0x55 : 0xAA;
  
  SetCheckSum();  
}

void KComfortFrame::SetOkValue(unsigned Value)
{
  SetOk();
  
  Data[8] = (unsigned char)(Value & 0xff);
  Data[9] = (unsigned char)(Value >> 8);
  
  SetCheckSum();
}

bool KComfortFrame::EqualData(const KComfortFrame & Frame) const
{
  for (int i = 0; i < 12; i++)
    if (Data[i] != Frame.Data[i])
      return false;
  
  return true;
}

void KComfortFrame::SetToString(char * Str) const
{
  Str[0] = 0;
  for (int i = 0; i < 12; i++)
    sprintf(Str + strlen(Str), "%2.2x", Data[i]);
}

bool KComfortFrame::SetFromString(const char * Str, bool AppendCheck)
{
  for (int i = 0; i < 12; i++)
  {
    while (*Str == ' ' || *Str == '.') Str++;
    if (!NCanUtils::IsHexDigit(Str[0]) || !NCanUtils::IsHexDigit(Str[1]))
    {
      if (!AppendCheck)
        return false;
      
      if (i != 10)
        return false;
      
      SetCheckSum();
      return true;
    }
    Data[i] = 16*NCanUtils::GetHexDigit(Str[0]) +
                 NCanUtils::GetHexDigit(Str[1]);
    Str += 2;
  }
  
  return true;
}

bool KComfortFrame::IsInitFrame() const
{
  return EqualData(InitFrame1) ||
         EqualData(InitFrame2);
}

void KComfortFrame::SetInitOk()
{
  SetOkValue(0x089c);
}


KCan232Frame::KCan232Frame(const char * str)
{
  *this = str;
}

bool KCan232Frame::SetCanFrame(KCanFrame & Frame) const
{
  Frame.Init();
  
  const char * Ptr = msg + 1;
  unsigned char b[4];
  
  if (!NCanUtils::GetHexByte(Ptr, b[0]))
    return false;
  
  switch (msg[0])
  {
    case 't':
    case 'r':
      if (b[0] > 0x7f ||
          !NCanUtils::IsHexDigit(*Ptr))
        return false;
    
      Frame.Id = 16*b[0] + NCanUtils::GetHexDigit(*Ptr++);
      break;
      
    case 'T':
    case 'R':
      if (b[0] > 0x3f ||
          !NCanUtils::GetHexByte(Ptr, b[1]) ||
          !NCanUtils::GetHexByte(Ptr, b[2]) ||
          !NCanUtils::GetHexByte(Ptr, b[3]))
        return false;
    
      for (int i = 0; i < 4; i++)
        Frame.Id = 256*Frame.Id + b[i];
      Frame.Id |= 0x80000000;
      break;
      
    default:
      return false;
  }
  if (msg[0] == 'r' || msg[0] == 'R') // RTR
    Frame.Id |= 0x40000000;
  
  if (*Ptr < '0' || *Ptr > '8')
    return false;
  
  Frame.Len = *Ptr++ - '0';
  
  for (int i = 0; i < Frame.Len; i++)
    if (!NCanUtils::GetHexByte(Ptr, Frame.Data[i]))
      return false;
  
  return true;  
}

bool KCan232Frame::IsOk() const
{
  KCanFrame Frame;
  
  return SetCanFrame(Frame);
}

bool KCan232Frame::GetCanFrame(const KCanFrame & Frame)
{
  if (Frame.Id & 0x80000000)
    sprintf(msg, "T%8.8X%1.1X", Frame.Id & 0x3ffffff, Frame.Len & 0xf);
  else
    sprintf(msg, "t%3.3X%1.1X", Frame.Id & 0x7ff, Frame.Len & 0xf);
 
  // RTR
  if (Frame.Id & 0x40000000)
    msg[0] += 'R' - 'T';
  
  for (int i = 0; i < Frame.Len; i++)
    sprintf(msg + strlen(msg), "%2.2X", Frame.Data[i]);
  
  return true;
}

KCan232Frame & KCan232Frame::operator =(const char * str)
{
  unsigned l = 0;
  int m = 0;
  while (str && str[m] && l < High(msg))
  {
    if (str[m] > ' ')
      msg[l++] = str[m];
    m++;
  }
  msg[l] = 0;
  
  return *this;
}

